import tensorflow as tf
from medgeconv import util_disjoint


class TestTFFunctions(tf.test.TestCase):
    def test_pdist(self):
        points = tf.constant([
            [[0, 0], [0, 1]],
            [[1, 1], [4, 5]],
            [[-1, -1], [-1, -1]],
        ], dtype="float32")

        target = tf.constant([
            [[0, 1], [1, 0]],
            [[0, 5], [5, 0]],
            [[0, 0], [0, 0]],
        ], dtype="float32")
        result = util_disjoint.pdist(points)
        self.assertAllClose(result, target)

    def test_get_knn_from_points(self):
        points = tf.constant([
            [0, 0], [0, 3], [0, 2],
            [1, 1], [4, 5], [5, 5], [1, 1.1],
        ], dtype="float32")
        is_valid = tf.constant([
            [1, 1, 1, 0],
            [1, 1, 1, 1]
        ], dtype="int32")
        target = tf.constant([
            [2, 1], [2, 0], [1, 0],
            [6, 4], [5, 6], [4, 6], [3, 4],
        ], dtype="int32")
        result = util_disjoint.get_knn_from_disjoint(points, k=2, is_valid=is_valid)
        self.assertAllEqual(result, target)

    def test_reduce_mean_valid(self):
        points = tf.constant([
            [1, 2], [3, 4],
            [9, 10], [11, 12], [13, 14], [15, 16],
        ], dtype="float32")
        is_valid = tf.constant([
            [1, 1, 0, 0],
            [1, 1, 1, 1]
        ], dtype="int32")
        target = tf.constant([
            [2, 3],
            [12, 13],
        ], dtype="float32")
        result = util_disjoint.reduce_mean_valid_disjoint(
            points, is_valid=is_valid)
        self.assertAllClose(result, target)

    def test_get_graph_ids(self):
        is_valid = tf.constant([
            [1, 1, 0, 0],
            [1, 1, 1, 1]
        ], dtype="int32")
        graph_ids = util_disjoint.get_graph_ids(is_valid)

        target = tf.constant([0, 0, 1, 1, 1, 1], dtype="int32")
        self.assertAllEqual(graph_ids, target)



import os
import tempfile
import tensorflow as tf
import numpy as np
import medgeconv


class TestEdgy(tf.test.TestCase):
    def setUp(self):
        self.n_points, self.n_features, self.n_coords = 5, 4, 3
        self.batchsize = 2
        self.units = [5, 2]
        self.next_neighbors = 3

        inp_points = tf.keras.layers.Input((self.n_points, self.n_features), batch_size=self.batchsize)
        inp_valid = tf.keras.layers.Input((self.n_points, ), batch_size=self.batchsize)
        inp_coords = tf.keras.layers.Input((self.n_points, self.n_coords), batch_size=self.batchsize)
        self.inps = (inp_points, inp_valid, inp_coords)
        self.edge_layer = medgeconv.DisjointEdgeConvBlock(
            units=self.units,
            next_neighbors=self.next_neighbors,
            kernel_initializer="ones",
            to_disjoint=True
        )
        self.edge_out = self.edge_layer(self.inps)[0]
        self.model = tf.keras.Model(self.inps, self.edge_out)

        points = np.arange(
            self.batchsize * self.n_points * self.n_features).reshape(
            *self.model.input_shape[0]).astype("float32")
        is_valid = np.ones((self.batchsize, self.n_points))
        is_valid[:, -1] = 0
        coords = points[:, :, :self.n_coords]
        self.x = points, is_valid, coords
        self.y = np.zeros((int(is_valid.sum()), self.units[-1]))

    def test_output_shape(self):
        self.assertTupleEqual(
            self.model.output_shape, (None, self.units[-1]))

    def test_output(self):
        points = np.arange(
            self.batchsize * self.n_points * self.n_features).reshape(
            *self.model.input_shape[0]).astype("float32")
        is_valid = np.ones((self.batchsize, self.n_points))
        is_valid[:, -1] = 0
        coords = points[:, :, :self.n_coords]

        output = self.model.predict((points, is_valid, coords))
        target = np.array([
            [5.9970026,   5.9970026],
            [95.249084,  95.249084],
            [281.07126, 281.07126],
            [483.5435, 483.5435],

            [355.6873, 355.6873],
            [558.15955, 558.15955],
            [760.6317, 760.6317],
            [963.10394, 963.10394]
        ], dtype="float32")
        np.testing.assert_almost_equal(target, output, decimal=4)

    def test_train(self):
        self.model.compile("sgd", "mse")
        self.model.train_on_batch(x=self.x, y=self.y)

    def test_loading(self):
        self.model.compile("sgd", "mse")
        with tempfile.TemporaryDirectory() as tempdir:
            path = os.path.join(tempdir, "temp.h5")
            tf.keras.models.save_model(self.model, path)
            loaded = tf.keras.models.load_model(
                path, custom_objects=medgeconv.custom_objects)
            loaded.train_on_batch(x=self.x, y=self.y)

    def test_with_pooling(self):
        edge_out = medgeconv.DisjointEdgeConvBlock(
            units=self.units,
            next_neighbors=self.next_neighbors,
            kernel_initializer="ones",
            to_disjoint=True,
            pooling=True,
        )(self.inps)
        model = tf.keras.Model(self.inps, edge_out)
        output = model.predict_on_batch(self.x)
        target = np.array([
            [216.46521, 216.46521],
            [659.3956, 659.3956]], dtype="float32")
        np.testing.assert_almost_equal(target, output)



"""
Keras layers for graph functionality.

"""
import tensorflow.keras as ks
import tensorflow.keras.layers as layers
from . import util


class EdgeConv:
    """
    The EdgeConv layer like it is used in ParticleNet, implemented
    as a pseudo-block.
    Has a dense network as the kernel network.

    Input: (nodes, is_valid, coordinates)

    nodes, shape (batchsize, n_nodes_max, n_features)
        Node features of the graph, padded to fixed size.
    is_valid, shape (batchsize, n_nodes_max)
        1 for actual node, 0 for padded node.
    coordinates, shape (batchsize, n_nodes_max, n_coords)
        Features of each node used for calculating nearest
        neighbors.

    Output, shape (batchsize, n_nodes_max, units[-1])

    Parameters
    ----------
    units : List or int
        How many dense layers are in the kernel network, and how many
        neurons does each of them have? E.g. [64, 64] means two dense
        layers with 64 neurons each.
    next_neighbors : int
        How many next neighbors to construct the edge features with for
        each node.
    kernel_initializer : str
    activation : str
    shortcut : bool
        Add a shortcut connection between input and output?

    """
    def __init__(self, units,
                 next_neighbors=16,
                 kernel_initializer="glorot_uniform",
                 activation="relu",
                 shortcut=True):
        self.units = units
        self.next_neighbors = next_neighbors
        self.kernel_initializer = kernel_initializer
        self.activation = activation
        self.shortcut = shortcut

        self.kernel_network = lambda xi_xj: kernel_nn(
            xi_xj, units=self.units,
            kernel_initializer=self.kernel_initializer,
            activation=self.activation)

    def __call__(self, inputs):
        nodes, is_valid, coordinates = inputs

        # get central and neighbour point features for each edge
        # between neighbours, defined by coordinates
        xi_xj = GetEdgeFeatures(
            next_neighbors=self.next_neighbors,
        )((nodes, is_valid, coordinates))

        x = self.get_edges(xi_xj)
        # remove padded nodes
        x, rev = util.flatten_graphs(x, is_valid)
        # apply MLP on points of each edge
        x = self.kernel_network(x)
        # aggregate over the nearest neighbours
        x = ks.backend.mean(x, axis=-2)

        if self.shortcut:
            sc, _ = util.flatten_graphs(nodes, is_valid)
            sc = ks.layers.Dense(
                self.units[-1], use_bias=False,
                kernel_initializer=self.kernel_initializer)(sc)
            sc = ks.layers.BatchNormalization()(sc)
            x = ks.layers.Activation(self.activation)(sc + x)

        x = rev(x)
        return x

    def get_edges(self, xi_xj):
        """
        Get the edge features, derived from the central point and
        the neighboring points.

        Output shape: (bs, n_points, n_neighbors, 2*n_features)

        """
        xi, xj = xi_xj
        dif = layers.Subtract()([xi, xj])
        x = layers.Concatenate(axis=-1)([xi, dif])
        return x


def kernel_nn(x, units, activation="relu", kernel_initializer='glorot_uniform'):
    """ The kernel network used on each edge of the graph. """
    if isinstance(units, int):
        units = [units]
    for uts in units:
        x = layers.Dense(uts, use_bias=False, activation=None,
                         kernel_initializer=kernel_initializer)(x)
        x = layers.BatchNormalization()(x)
        x = layers.Activation(activation)(x)
    return x


class GetEdgeFeatures(ks.layers.Layer):
    """
    Get the edge features of the graph.

    Input shape:
        [(bs, n_points, n_features),
         (bs, n_points),
         (bs, n_points, n_coordinates)]
        --> [nodes, is_valid, coordinates]
        - all the nodes of the graph, together with their features
        - Info about which nodes are valid nodes
        - Coordinates that will be used for calculateing the distance / knn

    Output shape:
        [
         (bs, n_points, next_neighbors, n_features),
         (bs, n_points, next_neighbors, n_features),
        ]
        --> [xi, xj]
        These matrices represent the edges of the graph:
        for [, i, j]:
        matrix 1 is xi (the central point),
        and matrix 2 is xj (the nearest neighbor j of central point i)

    """
    def __init__(self, next_neighbors, **kwargs):
        super().__init__(**kwargs)
        self.next_neighbors = next_neighbors

    def call(self, inputs):
        points, is_valid, coordinates = inputs
        # get the k nearest neighbours for each node (dont connect valid and invalid nodes)
        knn = util.get_knn_from_points(
            coordinates,
            self.next_neighbors,
            is_valid=is_valid,
        )
        # get central and neighbour point for each edge between neighbours
        xi_xj = util.get_xixj_knn(points, knn=knn, k=self.next_neighbors)
        return xi_xj

    def compute_output_shape(self, input_shape):
        return (input_shape[0, :2] + (self.next_neighbors,) + input_shape[0, -1:], ) * 2

    def get_config(self):
        config = super().get_config()
        config["next_neighbors"] = self.next_neighbors
        return config


class GlobalAvgValidPooling(ks.layers.Layer):
    """
    An average valid pooling layer.
    Pools a graph by averaging features over all nodes (or only all
    valid nodes).

    Input shape:    [(bs, n_points, n_features), (bs, n_points)]
    Output shape:   (bs, n_features)

    """
    def __init__(self, divide_by_nodes=True, **kwargs):
        super().__init__(**kwargs)
        self.divide_by_nodes = divide_by_nodes

    def call(self, inputs):
        points, is_valid = inputs
        return util.reduce_mean_valid(
            points, is_valid, divide_by_nodes=self.divide_by_nodes)

    def compute_output_shape(self, input_shape):
        return input_shape[:-2] + input_shape[-1:]

    def get_config(self):
        config = super().get_config()
        config["divide_by_nodes"] = self.divide_by_nodes
        return config



import tensorflow as tf


def get_knn_from_disjoint(nodes, k, is_valid):
    """
    Get the k nearest neighbors depending on distance.

    Parameters
    ----------
    nodes : tf.Tensor
        shape (None, n_features)
    k : int
        Number of nearest neighbors (excluding self).
    is_valid : tf.Tensor
        int32 tensor, shape (bs, n_points).

    Returns
    -------
    shape (None, k)
        int32, for each point, the indices of the points that are the
        nearest neighbors.

    """
    def func(args):
        graph = args
        euc_dist = pdist(graph, single_mode=True, take_sqrt=False)
        return tf.math.top_k(-euc_dist, k=k + 1)[1][:, 1:]

    n_valid_nodes = tf.reduce_sum(is_valid, axis=-1, keepdims=True)

    # Shape (batchsize, None, n_features)
    nodes_ragged = tf.RaggedTensor.from_row_lengths(
        nodes, tf.squeeze(n_valid_nodes), validate=False)
    # shape (batchsize, None, k)
    knn_ragged = tf.map_fn(
        func,
        nodes_ragged,
        fn_output_signature=tf.RaggedTensorSpec(
            shape=(None, k),
            dtype="int32",
            ragged_rank=0,
            row_splits_dtype=tf.dtypes.int32)
    )
    # shape (None, k)
    knn = knn_ragged.merge_dims(0, 1)

    graph_indices = tf.gather_nd(
        is_valid * tf.cumsum(n_valid_nodes, exclusive=True),
        tf.where(is_valid == 1))

    return knn + tf.expand_dims(graph_indices, -1)


def pdist(points, take_sqrt=True, single_mode=False):
    """
    Computes pairwise distance between each pair of points

    Parameters
    ----------
    points : tf.Tensor
        [bs, N, D] matrix representing a batch of N D-dimensional vectors each
    take_sqrt : bool
        Take the sqrt of the squared distance? Expensive...
    single_mode : bool
        If true, this is the operation for a single sample, not the batch

    Returns
    -------
    [bs, N, N] matrix of (squared) euclidean distances

    """
    if single_mode:
        transp_axes = None
    else:
        transp_axes = [0, 2, 1]

    x2 = tf.reduce_sum(tf.square(points), axis=-1, keepdims=True)
    cross = tf.matmul(
      points,
      tf.transpose(points, transp_axes)
    )
    output = x2 - 2 * cross + tf.transpose(x2, transp_axes)

    if take_sqrt:
        return tf.sqrt(output)
    else:
        return output


def get_xixj_disjoint(nodes, knn, k):
    """
    Get the features of each edge in the graph.

    Paramters
    ---------
    nodes : tf.Tensor
        shape (None, n_features)
    knn : tf.Tensor
        shape (None, k)
        int32, for each point, the indices of the points that are the
        nearest neighbors.
    k : int
        Number of nearest neighbors (excluding self).

    Returns
    -------
    tuple
        Two Tensors with shape (None, k, n_features).
        --> [?, i, j] desribes the edge between points i and j.
        The first matrix is xi, i.e. for each edge the central point
        The 2nd matrix is xj, i.e. for each edge the other point.

    """
    nodes_central = tf.tile(
        tf.expand_dims(nodes, axis=-2),
        [1, k, 1]
    )

    # TODO this produces a 'Converting sparse IndexedSlices to a dense Tensor
    #  of unknown shape.' warning. Thats because nodes has an unknown shape
    #  (None, n_features), along first axis is gathered.
    nodes_neighbors = tf.gather(nodes, knn)

    return nodes_central, nodes_neighbors


def reduce_mean_valid_disjoint(nodes, is_valid):
    """
    Average over valid nodes.

    Parameters
    ----------
    nodes : tf.Tensor
        shape (None, n_features)
    is_valid : tf.Tensor
        int32 tensor shape (bs, n_points).

    Returns
    -------
    shape (bs, n_features)
        For each feature, aggeregated over all valid nodes.

    """
    graph_ids = get_graph_ids(is_valid)
    pooled = tf.math.segment_mean(nodes, graph_ids)
    pooled.set_shape(is_valid.shape[:1] + pooled.shape[1:])
    return pooled


def get_graph_ids(is_valid):
    """ Shape (None,). """
    return tf.gather_nd(
        is_valid * tf.expand_dims(tf.range(is_valid.shape[0]), -1),
        tf.where(is_valid == 1))



"""
Tensorflow functions for graph functionality.

"""
import tensorflow as tf


def get_knn_from_points(points, k, is_valid):
    """
    Get the k nearest neighbors depending on distance.

    Parameters
    ----------
    points : tf.Tensor
        shape (bs, n_points, n_features)
    k : int
        Number of nearest neighbors (excluding self).
    is_valid : tf.Tensor
        boolean or float tensor shape (bs, n_points).

    Returns
    -------
    shape (bs, n_points, k), int32
        The index of which points are the nearest neighbors.

    """
    def func(args):
        return _get_knn_map(args, k=k)
    return tf.map_fn(func, [points, is_valid], dtype="int32")


def _get_knn_map(args, k):
    """ Get knn for one graph at a time. Only operates on real nodes. """
    graph, val = args
    # graph: (n_points, n_features), val (n_points, )

    # dummy tensor of fixed shape (n_points, k)
    knn_base = tf.zeros_like(val, dtype="int32")
    knn_base = tf.tile(tf.expand_dims(knn_base, -1), (1, k))

    # take only real nodes
    n_points_real = tf.cast(tf.reduce_sum(val), "int32")
    graph_prop = graph[:n_points_real]
    # get knn
    euc_dist = pdist(graph_prop, single_mode=True)  # (n_points_real, n_points_real)
    knn = tf.math.top_k(-euc_dist, k=k + 1)[1][:, 1:]  # (n_points_real, k)

    # insert into dummy tensor
    return tf.tensor_scatter_nd_update(
        knn_base, tf.expand_dims(tf.range(tf.shape(knn)[0]), -1), knn)


def get_xixj_knn(points, knn, k):
    """
    Given points of shape (bs, n_points, n_dims), get
    two matrices with shape (bs, n_points, k, n_dims).

    --> [?, i, j] desribes the edge between points i and j.
    The first matrix is xi, i.e. for each edge the central point
    The 2nd matrix is xj, i.e. for each edge the other point.

    """
    # TODO set invalid nodes to 0!
    point_central = tf.tile(
        tf.expand_dims(points, axis=-2),
        [1, 1, k, 1]
    )
    point_cloud_neighbors = tf.gather(points, knn, batch_dims=1)

    return point_central, point_cloud_neighbors


def pdist(points, take_sqrt=True, single_mode=False):
    """
    Computes pairwise distance between each pair of points

    Parameters
    ----------
    points : tf.Tensor
        [bs, N, D] matrix representing a batch of N D-dimensional vectors each
    take_sqrt : bool
        Take the sqrt of the squared distance? Expensive...
    single_mode : bool
        If true, this is the operation for a single sample, not the batch

    Returns
    -------
    [bs, N, N] matrix of (squared) euclidean distances

    """
    if single_mode:
        transp_axes = None
    else:
        transp_axes = [0, 2, 1]

    x2 = tf.reduce_sum(tf.square(points), axis=-1, keepdims=True)
    cross = tf.matmul(
      points,
      tf.transpose(points, transp_axes)
    )
    output = x2 - 2 * cross + tf.transpose(x2, transp_axes)

    if take_sqrt:
        return tf.sqrt(output)
    else:
        return output


def reduce_mean_valid(points, is_valid, divide_by_nodes=True):
    """
    Sum up or average over valid nodes.

    Parameters
    ----------
    points : tf.Tensor
        shape (bs, n_points, n_features)
    is_valid : tf.Tensor
        boolean or float tensor shape (bs, n_points).
    divide_by_nodes : bool
        If true, divide by number of valid nodes (= average).

    Returns
    -------
    shape (bs, n_features)
        For each feature, aggeregated over all valid nodes.

    """
    is_valid = tf.cast(is_valid, points.dtype)
    # set node features of invalid nodes to 0
    valid_points = points * tf.expand_dims(is_valid, -1)
    if divide_by_nodes:
        # number of valid nodes in each batch
        n_valid_nodes = tf.reduce_sum(is_valid, axis=-1, keepdims=True)
        # sum over all nodes
        summed_nodes = tf.reduce_sum(valid_points, axis=-2)
        # divide the sum by the number of valid nodes per feature
        return summed_nodes / tf.maximum(n_valid_nodes, tf.keras.backend.epsilon())
    else:
        return tf.reduce_mean(valid_points, axis=1)


def reduce_mean_valid_disjoint(points_disjoint, is_valid):
    """
    Average over valid nodes.

    Parameters
    ----------
    points_disjoint : tf.Tensor
        shape (None, n_features)
    is_valid : tf.Tensor
        boolean or float tensor shape (bs, n_points).

    Returns
    -------
    shape (bs, n_features)
        For each feature, aggeregated over all valid nodes.

    """
    is_valid = tf.cast(is_valid, points_disjoint.dtype)

    placeholder = tf.tile(
        tf.expand_dims(tf.zeros_like(is_valid), -1),
        (1, 1, points_disjoint.shape[-1])
    )
    points_dense = tf.tensor_scatter_nd_update(
        placeholder, tf.where(is_valid == 1), points_disjoint)

    # number of valid nodes in each batch
    n_valid_nodes = tf.reduce_sum(is_valid, axis=-1, keepdims=True)
    # sum over all nodes
    summed_nodes = tf.reduce_sum(points_dense, axis=-2)
    # divide the sum by the number of valid nodes per feature
    return summed_nodes / tf.maximum(n_valid_nodes, tf.keras.backend.epsilon())


def flatten_graphs(nodes, is_valid):
    """
    Flatten node and batch dimension and remove non-existing nodes from
    graph. Also reverse this easily.

    Parameters
    ----------
    nodes : tf.Tensor
        The node features, shape (bs, n_nodes, ...)
    is_valid : tf.Tensor
        Which nodes are valid nodes? Shape (bs, n_nodes)

    Returns
    -------
    flattened : tf.Tensor
        The flattened nodes, shape (?, ...)
    rev : function
        Apply this on the flattened tensor to get the original shape back
        (with padded zeros), i.e. turn it back to (bs, n_nodes, ...).
        First axis (the batchsize/graph axis) must have
        not changed shape or this wont work.

    """
    valid_indices = tf.where(is_valid == 1)
    flattened = tf.gather_nd(nodes, valid_indices)

    def reverse(flat_nodes):
        empty = tf.tile(
            tf.expand_dims(tf.zeros_like(is_valid), -1),
            (1, 1, flat_nodes.shape[-1])
        )
        return tf.tensor_scatter_nd_update(
            empty, valid_indices, flat_nodes)

    return flattened, reverse


def dense_to_ragged(points, is_valid):
    """ (batchsize, n_points, n_features) --> (batchsize, None, n_features) """
    n_valid_nodes = tf.reduce_sum(tf.cast(is_valid, "int32"), -1)
    n_invalid_nodes = tf.ones_like(n_valid_nodes) * points.shape[1] - n_valid_nodes
    row_lengths = tf.reshape(tf.stack([n_valid_nodes, n_invalid_nodes], -1), [-1])
    points_flat = tf.reshape(points, [-1, points.shape[-1]])
    ragged = tf.RaggedTensor.from_row_lengths(points_flat, row_lengths, validate=False)
    return ragged[::2]


def dense_to_sparse(points, is_valid):
    return tf.sparse.from_dense(points * tf.expand_dims(is_valid, -1))



import time
import numpy as np
import tensorflow as tf
import matplotlib.pyplot as plt
from medgeconv.layers import EdgeConv


NEXT_NEIGHBORS = 16
N_POINTS_MAX = 2000
N_FEATURES = 7
N_COORDS = 4
BATCHSIZE = 64
UNITS = (64, 64, 64)


def get_model():
    inps = (
        tf.keras.layers.Input((N_POINTS_MAX, N_FEATURES), batch_size=BATCHSIZE),
        tf.keras.layers.Input((N_POINTS_MAX,), batch_size=BATCHSIZE),
        tf.keras.layers.Input((N_POINTS_MAX, N_COORDS), batch_size=BATCHSIZE)
    )
    edge_layer = EdgeConv(
        units=UNITS,
        next_neighbors=NEXT_NEIGHBORS,
        kernel_initializer="ones",
    )
    edge_out = edge_layer(inps)
    model = tf.keras.Model(inps, edge_out)
    model.compile("sgd", "mse")
    return model


class DataGenerator(tf.keras.utils.Sequence):
    def __init__(self, length=1000, batches=10):
        self.length = length
        self.batches = batches
        self.x, self.y = get_data(self.batches)

    def __len__(self):
        return self.length

    def __getitem__(self, item):
        item = item % self.batches
        slc = slice(item*BATCHSIZE, (item+1)*BATCHSIZE)
        return [x_[slc] for x_ in self.x], self.y[slc]

    def plot_n_nodes(self):
        plt.hist(self.x[1].sum(-1), bins=50)
        plt.xlabel("valid nodes")


def get_data(batches=10):
    points = np.random.rand(batches*BATCHSIZE, N_POINTS_MAX, N_FEATURES).astype("float32")
    is_valid = np.zeros((batches*BATCHSIZE, N_POINTS_MAX)).astype("float32")
    coords = points[:, :, :N_COORDS]
    # number of valid nodes, linearily decreasing
    n_valid_nodes = np.random.triangular(
        NEXT_NEIGHBORS+1, NEXT_NEIGHBORS+1, N_POINTS_MAX+1, batches*BATCHSIZE).astype(int)
    # exponential
    # np.minimum(N_POINTS_MAX+1, np.random.exponential(300, batches*BATCHSIZE)).astype(int) + NEXT_NEIGHBORS+1

    for i, l in enumerate(n_valid_nodes):
        is_valid[i, :l] = 1.

    x = points, is_valid, coords
    y = np.zeros((batches*BATCHSIZE, N_POINTS_MAX, UNITS[-1]))
    return x, y


def benchmark(warmup_batches=10, profile_batches=10, n_loops=100, tensorboarding=True, plotit=True):
    """
    For profiling on GPU:
        (0.)  pip install -U tensorboard_plugin_profile
         1.   module load cuda/10.1
         2.   export LD_LIBRARY_PATH=/usr/local/cuda-10.1/extras/CUPTI/lib64/

    To display:
        1. firefox &
        2. tensorboard --logdir benchmarks/
        3. URL:  http://localhost:6006/

    """
    model = get_model()

    if tensorboarding:
        benchmark_dir = "benchmarks"
        print(f"TensorBoarding to '{benchmark_dir}'")
        tensorboard = tf.keras.callbacks.TensorBoard(
            log_dir=benchmark_dir,
            profile_batch=f"{warmup_batches+1}, {warmup_batches+profile_batches}",
            histogram_freq=1,
            write_images=True,
        )
        generator = DataGenerator(length=warmup_batches+profile_batches+1)
        val_data = generator[warmup_batches+profile_batches+1]
        model.fit(
            x=generator,
            validation_data=val_data,
            verbose=1,
            callbacks=[tensorboard],
            steps_per_epoch=warmup_batches+profile_batches,
        )
        print("tensorboarding complete")

    times = np.zeros(n_loops)
    generator = DataGenerator(length=n_loops)
    for i in range(n_loops):
        if i % 10 == 0:
            print(f"Batch {i}/{n_loops}...")
        x, y = generator[i]
        start_time = time.time()
        model.fit(x, y, verbose=0)
        times[i] = time.time() - start_time
    print("Done! Times excluding first batch:")
    print(f"Total time: {times[1:].sum():.4f}")
    print(f"Per batch: {times[1:].mean():.4f}")
    print(f"   median: {np.median(times[1:]):.4f}")
    if plotit:
        plt.plot(times[1:])
    return times


def _get_inp(batch_size=2):
    points = tf.keras.Input((5, 2), batch_size=batch_size)
    is_valid = tf.keras.Input((5, ), batch_size=batch_size)
    coords = tf.keras.Input((5, 2), batch_size=batch_size)
    return points, is_valid, coords


def _get_inp_numpy(batch_size=2):
    points = np.arange(0, batch_size*5*2, dtype="float32").reshape((batch_size, 5, 2))
    is_valid = np.array([[1, 1, 1, 1, 1], [1, 1, 1, 0, 0]], dtype="float32")
    coords = points
    return points, is_valid, coords


def _get_inp_const(batch_size=2):
    return [tf.constant(x) for x in _get_inp_numpy(batch_size)]


if __name__ == '__main__':
    benchmark(tensorboarding=False, plotit=False)




import tensorflow as tf
from . import layers, layers_disjoint


class DisjointEdgeConvBlock(layers_disjoint.DisjointEdgeConv):
    """
    DisjointEdgeConv with additional options as used in the ParticleNet
    paper.

    Parameters
    ----------
    to_disjoint : bool
        Start by transforming dense input to disjoint.
        Has to be used if this is the first layer.
    batchnorm_for_nodes : bool
        Use batchnorm layers for node features?
        Recommended for first layer.
    pooling : bool
        Finish with average pooling?
        Has to be used if this is the last layer.

    """
    def __init__(self, units,
                 next_neighbors,
                 kernel_initializer="glorot_uniform",
                 activation="relu",
                 shortcut=True,
                 to_disjoint=False,
                 batchnorm_for_nodes=False,
                 pooling=False):
        super().__init__(
            units=units,
            next_neighbors=next_neighbors,
            kernel_initializer=kernel_initializer,
            activation=activation,
            shortcut=shortcut,
        )
        self.to_disjoint = to_disjoint
        self.batchnorm_for_nodes = batchnorm_for_nodes
        self.pooling = pooling

    def __call__(self, x):
        nodes, is_valid, coordinates = x

        if self.to_disjoint:
            nodes, is_valid, coordinates = layers_disjoint.DenseToDisjoint()(
                (nodes, is_valid, coordinates))

        if self.batchnorm_for_nodes:
            nodes = tf.keras.layers.BatchNormalization()(nodes)

        nodes = super().__call__((nodes, is_valid, coordinates))

        if self.pooling:
            return layers_disjoint.GlobalAvgPoolingDisjoint()((nodes, is_valid))
        else:
            return nodes, is_valid, nodes


custom_objects = {
    # m:
    "GlobalAvgValidPooling": layers.GlobalAvgValidPooling,
    "GetEdgeFeatures": layers.GetEdgeFeatures,
    # disjoint:
    "GlobalAvgPoolingDisjoint": layers_disjoint.GlobalAvgPoolingDisjoint,
    "GetEdgeFeaturesDisjoint": layers_disjoint.GetEdgeFeaturesDisjoint,
    "DenseToDisjoint": layers_disjoint.DenseToDisjoint,
}



import tensorflow as tf
from medgeconv import util


class TestTFFunctions(tf.test.TestCase):
    def test_pdist(self):
        points = tf.constant([
            [[0, 0], [0, 1]],
            [[1, 1], [4, 5]],
            [[-1, -1], [-1, -1]],
        ], dtype="float32")

        target = tf.constant([
            [[0, 1], [1, 0]],
            [[0, 5], [5, 0]],
            [[0, 0], [0, 0]],
        ], dtype="float32")
        result = util.pdist(points)
        self.assertAllClose(result, target)

    def test_get_knn_from_points(self):
        points = tf.constant([
            [[0, 0], [0, 3], [0, 2], [0, 1]],
            [[1, 1], [4, 5], [5, 5], [1, 1.1]],
        ], dtype="float32")
        is_valid = tf.constant([
            [1, 1, 1, 0],
            [1, 1, 1, 1]
        ], dtype="float32")
        target = tf.constant([
            [[2, 1], [2, 0], [1, 0], [0, 0]],
            [[3, 1], [2, 3], [1, 3], [0, 1]],
        ], dtype="float32")
        result = util.get_knn_from_points(points, k=2, is_valid=is_valid)
        self.assertAllClose(result, target)

    def test_reduce_mean_valid(self):
        points = tf.constant([
            [[1, 2], [3, 4], [5, 6], [7, 8]],
            [[9, 10], [11, 12], [13, 14], [15, 16]],
        ], dtype="float32")
        is_valid = tf.constant([
            [1, 1, 0, 0],
            [1, 1, 1, 1]
        ], dtype="float32")
        target = tf.constant([
            [2, 3],
            [12, 13],
        ], dtype="float32")
        result = util.reduce_mean_valid(
            points, is_valid=is_valid, divide_by_nodes=True)
        self.assertAllClose(result, target)

    def test_flatten_graphs(self):
        points = tf.constant([
            [[1, 2], [3, 4], [0, 0], [0, 0]],
            [[9, 10], [11, 12], [13, 14], [15, 16]],
        ], dtype="float32")
        is_valid = tf.constant([
            [1, 1, 0, 0],
            [1, 1, 1, 1]
        ], dtype="float32")
        flattened, rev = util.flatten_graphs(nodes=points, is_valid=is_valid)

        target_flattened = tf.constant([
            [1, 2], [3, 4], [9, 10], [11, 12], [13, 14], [15, 16]
        ])
        self.assertAllClose(flattened, target_flattened)
        self.assertAllClose(rev(flattened), points)



#!/usr/bin/env python
from setuptools import setup

with open('requirements.txt') as fobj:
    requirements = [l.strip() for l in fobj.readlines()]

with open("README.rst") as fh:
    long_description = fh.read()

setup(
    name='medgeconv',
    author='Stefan Reck',
    author_email='stefan.reck@fau.de',
    long_description=long_description,
    setup_requires=['setuptools_scm'],
    use_scm_version=True,
    install_requires=requirements,
    packages=["medgeconv"],
)
